@@ -7,6 +7,7 @@ showcasing features of the channel object:
 - setting a callback for receiving channel data
 
 """
+
 import execnet
 
 gw = execnet.makegateway()
@@ -26,7 +27,7 @@ def write(data):
     print("received:", repr(data))
 
 
-outchan.setcallback(write)
+outchan.setcallback(write)  # type: ignore[attr-defined]
 
 gw.remote_exec(
     """
