@@ -5,13 +5,13 @@ obtain system info from remote machine.
 
 (c) Holger Krekel, MIT license
 """
+
 import optparse
 import re
 import sys
 
 import execnet
 
-
 parser = optparse.OptionParser(usage=__doc__)
 parser.add_option(
     "-f",
@@ -129,7 +129,7 @@ def getinfo(sshname, ssh_config=None, loginfo=sys.stdout):
     if ssh_config:
         spec = f"ssh=-F {ssh_config} {sshname}"
     else:
-        spec += "ssh=%s" % sshname
+        spec = "ssh=%s" % sshname
     debug("connecting to", repr(spec))
     try:
         gw = execnet.makegateway(spec)
