@@ -1,4 +1,4 @@
-
+from __future__ import annotations
 
 import execnet
 
@@ -11,20 +11,22 @@ def process_item(channel):
     # task processor, sits on each CPU
     import time
     import random
+
     channel.send("ready")
     for x in channel:
         if x is None:  # we can shutdown
             break
         # sleep random time, send result
         time.sleep(random.randrange(3))
-        channel.send(x*10)
+        channel.send(x * 10)
+
 
 # execute taskprocessor everywhere
 mch = group.remote_exec(process_item)
 
 # get a queue that gives us results
 q = mch.make_receive_queue(endmarker=-1)
-tasks = range(10)  # a list of tasks, here just integers
+tasks: list[int] | None = list(range(10))  # a list of tasks, here just integers
 terminated = 0
 while 1:
     channel, item = q.get()
@@ -36,14 +38,14 @@ while 1:
             break
         continue
     if item != "ready":
-        print("other side {} returned {!r}".format(channel.gateway.id, item))
-    if not tasks:
+        print(f"other side {channel.gateway.id} returned {item!r}")
+    if not tasks and tasks is not None:
         print("no tasks remain, sending termination request to all")
         mch.send_each(None)
-        tasks = -1
-    if tasks and tasks != -1:
+        tasks = None
+    if tasks:
         task = tasks.pop()
         channel.send(task)
-        print("sent task {!r} to {}".format(task, channel.gateway.id))
+        print(f"sent task {task!r} to {channel.gateway.id}")
 
 group.terminate()
